@@ -18,7 +18,6 @@
 #include <unistd.h>
 #include <sys/stat.h>
 #include <sys/statvfs.h>
-#include <sys/types.h>
 #include <sys/ioctl.h>
 #include <sys/mount.h>
 #include <stdio.h>
@@ -62,7 +61,7 @@ extern void print_readable_size_str(unsigned long long size_byte, char *new_size
 
     float new_size = 1.0;
     memset(new_size_str, 0, 11);
-    uint64_t tbyte=1000000000000;
+    uint64_t tbyte=1000000000000.0;
     uint64_t gbyte=1000000000;
     uint64_t mbyte=1000000;
     uint64_t kbyte=1000;
@@ -101,12 +100,15 @@ extern void usage(void)
             "\n"
             "    -o,  --output FILE      Output FILE\n"
             "    -O   --overwrite FILE   Output FILE, overwriting if exists\n"
+            "         --restore_row_file create special row file for loop device\n"
             "    -s,  --source FILE      Source FILE\n"
             "    -L,  --logfile FILE     Log FILE\n"
 #ifndef	RESTORE
             "    -c,  --clone            Save to the special image format\n"
             "    -r,  --restore          Restore from the special image format\n"
             "    -b,  --dev-to-dev       Local device to device copy mode\n"
+            "    -D,  --domain           Create ddrescue domain log from source device\n"
+            "         --offset_domain=X  Add offset X (bytes) to domain log values\n"
             "    -R,  --rescue           Continue after disk read errors\n"
 #endif
             "    -dX, --debug=X          Set the debug level to X = [0|1|2]\n"
@@ -119,6 +121,8 @@ extern void usage(void)
 	    "         --ignore_crc       Ignore crc check error\n"
             "    -F,  --force            Force progress\n"
             "    -f,  --UI-fresh         Fresh times of progress\n"
+            "    -m,  --max_block_cache  The used block will be cache until max number\n"
+            "    -q,  --quiet		 Disable progress message\n"
             "    -v,  --version          Display partclone version\n"
             "    -h,  --help             Display this help\n"
             , EXECNAME, VERSION, EXECNAME);
@@ -130,47 +134,58 @@ extern void print_version(void){
     exit(0);
 }
 
+enum {
+    OPT_OFFSET_DOMAIN = 1000
+};
+
 extern void parse_options(int argc, char **argv, cmd_opt* opt)
 {
-    static const char *sopt = "-hvd::L:cbro:O:s:f:RCXFINi";
+    static const char *sopt = "-hvd::L:cbrDo:O:s:f:m:RCXFINiql";
     static const struct option lopt[] = {
         { "help",		no_argument,	    NULL,   'h' },
         { "print_version",	no_argument,	    NULL,   'v' },
         { "output",		required_argument,  NULL,   'o' },
-        { "overwrite",	required_argument,  NULL,   'O' },
+        { "overwrite",		required_argument,  NULL,   'O' },
         { "source",		required_argument,  NULL,   's' },
         { "restore-image",	no_argument,	    NULL,   'r' },
+        { "restore_row_file",	no_argument,	    NULL,   'l' },
         { "clone-image",	no_argument,	    NULL,   'c' },
         { "dev-to-dev",		no_argument,	    NULL,   'b' },
+        { "domain",		no_argument,	    NULL,   'D' },
+        { "offset_domain",	required_argument,  NULL,   OPT_OFFSET_DOMAIN },
         { "debug",		optional_argument,  NULL,   'd' },
-        { "logfile",	required_argument,  NULL,   'L' },
+        { "logfile",		required_argument,  NULL,   'L' },
         { "rescue",		no_argument,	    NULL,   'R' },
-        { "UI-fresh",	required_argument,  NULL,   'u' },
+        { "UI-fresh",		required_argument,  NULL,   'f' },
+        { "max_block_cache",	required_argument,  NULL,   'm' },
         { "check",		no_argument,	    NULL,   'C' },
         { "dialog",		no_argument,	    NULL,   'X' },
-        { "ignore_fschk",   no_argument,    NULL,   'I' },
-	{ "ignore_crc",     no_argument,    NULL,   'i' },
+        { "ignore_fschk",	no_argument,	    NULL,   'I' },
+	{ "ignore_crc",		no_argument,	    NULL,   'i' },
         { "force",		no_argument,	    NULL,   'F' },
+        { "quiet",		no_argument,	    NULL,   'q' },
 #ifdef HAVE_LIBNCURSESW
         { "ncurses",		no_argument,	    NULL,   'N' },
 #endif
         { NULL,			0,		    NULL,    0  }
     };
 
-    char c;
+    int c;
     int mode = 0;
     memset(opt, 0, sizeof(cmd_opt));
     opt->debug = 0;
     opt->rescue = 0;
+    opt->max_block_cache = 8;
     opt->check = 1;
     opt->ignore_crc = 0;
+    opt->quiet = 0;
     opt->logfile = "/var/log/partclone.log";
 
 #ifdef RESTORE
     opt->restore++;
     mode++;
 #endif
-    while ((c = getopt_long(argc, argv, sopt, lopt, NULL)) != (char)-1) {
+    while ((c = getopt_long(argc, argv, sopt, lopt, NULL)) != -1) {
         switch (c) {
             case 's': 
                 opt->source = optarg;
@@ -201,12 +216,19 @@ extern void parse_options(int argc, char **argv, cmd_opt* opt)
                 opt->dd++;
                 mode++;
                 break;
+            case 'D':
+                opt->domain++;
+                mode++;
+                break;
             case 'd':
                 if (optarg)
                     opt->debug = atol(optarg);
                 else
                     opt->debug = 1;
                 break;
+            case 'm': 
+                opt->max_block_cache = atol(optarg);
+                break;
             case 'L': 
                 opt->logfile = optarg;
                 break;
@@ -222,7 +244,12 @@ extern void parse_options(int argc, char **argv, cmd_opt* opt)
             case 'i':
                 opt->ignore_crc = 1;
                 break;
-
+            case 'l':
+                opt->restore_row_file = 1;
+                break;
+            case 'q':
+                opt->quiet = 1;
+                break;
             case 'R':
                 opt->rescue++;
                 break;
@@ -240,6 +267,9 @@ extern void parse_options(int argc, char **argv, cmd_opt* opt)
             case 'C':
                 opt->check = 0;
                 break;
+            case OPT_OFFSET_DOMAIN:
+                opt->offset_domain = strtoull(optarg, NULL, 0);
+                break;
             default:
                 fprintf(stderr, "Unknown option '%s'.\n", argv[optind-1]);
                 usage();
@@ -271,10 +301,12 @@ extern void parse_options(int argc, char **argv, cmd_opt* opt)
         opt->source = "-";
     }
 
-    if (opt->clone){
+    if (opt->clone || opt->domain){
 
         if ((strcmp(opt->source, "-") == 0) || (opt->source == NULL)) {
-            fprintf(stderr, "Partclone can't clone from stdin.\nFor help, type: %s -h\n", EXECNAME);
+            fprintf(stderr, "Partclone can't %s from stdin.\nFor help, type: %s -h\n",
+                    opt->clone ? "clone" : "make domain log",
+                    EXECNAME);
             //usage();
             exit(0);
         }
@@ -634,7 +666,7 @@ extern int check_mem_size(image_head image_hdr, cmd_opt opt, unsigned long long
     void *test_mem;
 
     image_head_size = sizeof(image_head);
-    bitmap_size = (sizeof(char)*image_hdr.totalblock);
+    bitmap_size = sizeof(unsigned long)*LONGS(image_hdr.totalblock);
     crc_io_size = sizeof(unsigned long)+image_hdr.block_size;
     *mem_size = image_head_size + bitmap_size + crc_io_size;
 
@@ -649,32 +681,35 @@ extern int check_mem_size(image_head image_hdr, cmd_opt opt, unsigned long long
 }
 
 /// get bitmap from image file to restore data
-extern void get_image_bitmap(int* ret, cmd_opt opt, image_head image_hdr, char* bitmap){
-    int size, r_size;
-    int do_write = 0;
-    char* buffer;
-    unsigned long long block_id;
+extern void get_image_bitmap(int* ret, cmd_opt opt, image_head image_hdr, unsigned long* bitmap){
+    unsigned long long size, r_size, r_need;
+    char buffer[4096];
+    unsigned long long offset = 0;
     unsigned long long bused = 0, bfree = 0;
-    int debug = opt.debug;
+    int i, debug = opt.debug;
     int err_exit = 1;
 
     size = sizeof(char)*image_hdr.totalblock;
-    buffer = (char*)malloc(size);
 
-    r_size = read_all(ret, buffer, size, &opt);
-    memcpy(bitmap, buffer, size);
-
-    free(buffer);
-
-    for (block_id = 0; block_id < image_hdr.totalblock; block_id++){
-        if(bitmap[block_id] == 1){
-            //printf("u = %i\n",block_id);
-            bused++;
-        } else {
-            //printf("n = %i\n",block_id);
-            bfree++;
-        }
+    while (size > 0){
+	r_need = size > sizeof(buffer) ? sizeof(buffer) : size;
+	r_size = read_all(ret, buffer, r_need, &opt);
+	if (r_size < r_need){
+	    log_mesg(0, 1, 1, debug, "Unable to read bitmap.\n");
+	}
+	for (i = 0; i < r_need; i++){
+	    if(buffer[i] == 1){
+		pc_set_bit(offset + i, bitmap);
+		bused++;
+	    } else {
+		pc_clear_bit(offset + i, bitmap);
+		bfree++;
+	    }
+	}
+	offset += r_need;
+	size -= r_need;
     }
+
     if (debug >= 2) {
         if(image_hdr.usedblocks != bused){
             if (opt.force)
@@ -753,7 +788,7 @@ extern int open_source(char* source, cmd_opt* opt){
     int flags = O_RDONLY | O_LARGEFILE;
 
     log_mesg(1, 0, 0, debug, "open source file/device %s\n", source);
-    if((opt->clone) || (opt->dd)){ /// always is device, clone from device=source
+    if((opt->clone) || (opt->dd) || (opt->domain)){ /// always is device, clone from device=source
 
         mp = malloc(PATH_MAX + 1);
         if (check_mount(source, mp) == 1){
@@ -792,13 +827,13 @@ extern int open_target(char* target, cmd_opt* opt){
     struct stat st_dev;
 
     log_mesg(1, 0, 0, debug, "open target file/device\n");
-    if (opt->clone){
+    if (opt->clone || opt->domain){
         if (strcmp(target, "-") == 0){ 
             ret = fileno(stdout);
             if (ret == -1)
                 log_mesg(0, 1, 1, debug, "clone: open %s(stdout) error\n", target);
         } else { 
-            flags |= O_CREAT;		    /// new file
+            flags |= O_CREAT | O_TRUNC;	    /// new file
             if (!opt->overwrite)	    /// overwrite
                 flags |= O_EXCL;
             ret = open (target, flags, S_IRUSR);
@@ -841,11 +876,11 @@ extern int open_target(char* target, cmd_opt* opt){
 }
 
 /// the io function, reference from ntfsprogs(ntfsclone).
-extern int io_all(int *fd, char *buf, int count, int do_write, cmd_opt* opt)
+extern int io_all(int *fd, char *buf, unsigned long long count, int do_write, cmd_opt* opt)
 {
-    int i;
+    long long int i;
     int debug = opt->debug;
-    int size = count;
+    unsigned long long size = count;
 
     // for sync I/O buffer, when use stdin or pipe.
     while (count > 0) {
@@ -855,18 +890,18 @@ extern int io_all(int *fd, char *buf, int count, int do_write, cmd_opt* opt)
             i = read(*fd, buf, count);
 
         if (i < 0) {
-            log_mesg(0, 0, 1, debug, "%s: errno = %s(%i)\n",__func__, strerror(errno), errno);
+            log_mesg(1, 0, 1, debug, "%s: errno = %i(%s)\n",__func__, errno, strerror(errno));
             if (errno != EAGAIN && errno != EINTR){
                 return -1;
             }
         } else if (i == 0){
-            log_mesg(1, 0, 0, debug, "%s: errno = %s(%i)\n",__func__, strerror(errno), errno);
+            log_mesg(1, 0, 1, debug, "%s: nothing to read. errno = %i(%s)\n",__func__, errno, strerror(errno));
             return 0;
 
         } else {
             count -= i;
             buf = i + (char *) buf;
-            log_mesg(2, 0, 0, debug, "%s: read %li, %li left.\n",__func__, i, count);
+            log_mesg(2, 0, 0, debug, "%s: read %lli, %lli left.\n",__func__, i, count);
         }
     }
     return size;
@@ -948,6 +983,8 @@ extern void print_opt(cmd_opt opt){
         log_mesg(1, 0, 0, debug, "MODE: restore\n");
     else if (opt.dd)
         log_mesg(1, 0, 0, debug, "MODE: device to device\n");
+    else if (opt.domain)
+        log_mesg(1, 0, 0, debug, "MODE: create domain log for ddrescue\n");
 
     log_mesg(1, 0, 0, debug, "DEBUG: %i\n", opt.debug);
     log_mesg(1, 0, 0, debug, "SOURCE: %s\n", opt.source);
@@ -955,12 +992,15 @@ extern void print_opt(cmd_opt opt){
     log_mesg(1, 0, 0, debug, "OVERWRITE: %i\n", opt.overwrite);
     log_mesg(1, 0, 0, debug, "RESCUE: %i\n", opt.rescue);
     log_mesg(1, 0, 0, debug, "CHECK: %i\n", opt.check);
+    log_mesg(1, 0, 0, debug, "QUIET: %i\n", opt.quiet);
     log_mesg(1, 0, 0, debug, "FRESH: %i\n", opt.fresh);
     log_mesg(1, 0, 0, debug, "FORCE: %i\n", opt.force);
+    log_mesg(1, 0, 0, debug, "MAX BLOCK CACHE: %i\n", opt.max_block_cache);
 #ifdef HAVE_LIBNCURSESW
     log_mesg(1, 0, 0, debug, "NCURSES: %i\n", opt.ncurses);
 #endif
     log_mesg(1, 0, 0, debug, "DIALOG: %i\n", opt.dialog);
+    log_mesg(1, 0, 0, debug, "OFFSET DOMAIN: 0x%llX\n", opt.offset_domain);
 
 }
 
@@ -982,6 +1022,8 @@ extern void print_partclone_info(cmd_opt opt){
         log_mesg(0, 0, 1, debug, _("Starting to restore image (%s) to device (%s)\n"), opt.source, opt.target);
     else if(opt.dd)
         log_mesg(0, 0, 1, debug, _("Starting to back up device(%s) to device(%s)\n"), opt.source, opt.target);
+    else if (opt.domain)
+        log_mesg(0, 0, 1, debug, _("Starting to map device (%s) to domain log (%s)\n"), opt.source, opt.target);
     else
         log_mesg(0, 0, 1, debug, "unknow mode\n");
 }
@@ -1023,13 +1065,15 @@ extern void print_finish_info(cmd_opt opt){
     bindtextdomain(PACKAGE, LOCALEDIR);
     textdomain(PACKAGE);
     if (opt.chkimg)
-        log_mesg(0, 0, 1, debug, _("Partclone successfully checkd the image (%s)\n"), opt.source);	
+        log_mesg(0, 0, 1, debug, _("Partclone successfully checked the image (%s)\n"), opt.source);	
     else if (opt.clone)
         log_mesg(0, 0, 1, debug, _("Partclone successfully cloned the device (%s) to the image (%s)\n"), opt.source, opt.target);	
     else if(opt.restore)
         log_mesg(0, 0, 1, debug, _("Partclone successfully restored the image (%s) to the device (%s)\n"), opt.source, opt.target);
     else if(opt.dd)
         log_mesg(0, 0, 1, debug, _("Partclone successfully cloned the device (%s) to the device (%s)\n"), opt.source, opt.target);
+    else if (opt.domain)
+        log_mesg(0, 0, 1, debug, _("Partclone successfully mapped the device (%s) to the domain log (%s)\n"), opt.source, opt.target);
 
 }
 
@@ -1049,22 +1093,33 @@ static unsigned long long block_count(unsigned long long partition_size, int blo
 /// initial dd hdr
 extern void initial_dd_hdr(int ret, image_head* image_hdr){
 
-    memset(image_hdr ,(int )NULL, sizeof(image_head));
-    memcpy(image_hdr->fs, raw_MAGIC, FS_MAGIC_SIZE);
-    image_hdr->block_size  = 512*2;
+    memset(image_hdr, 0, sizeof(image_head));
+    strncpy(image_hdr->fs, raw_MAGIC, FS_MAGIC_SIZE);
+    image_hdr->block_size  = 512;
     image_hdr->device_size = get_partition_size(&ret);
     image_hdr->totalblock  = block_count(image_hdr->device_size, image_hdr->block_size);
     image_hdr->usedblocks  = image_hdr->totalblock;
 }
 
 /// initial bitmap
-extern void dd_bitmap(image_head image_hdr, char* bitmap){
+extern void dd_bitmap(image_head image_hdr, unsigned long* bitmap){
 
     int block;
 
     /// initial image bitmap as 1 (all block are used)
-    for(block = 0; block < image_hdr.totalblock; block++)
-        bitmap[block] = 1; 
+    memset(bitmap, 0xFF, sizeof(unsigned long)*LONGS(image_hdr.totalblock));
 
 }
 
+/// write last block
+void write_last_block(int* dfw, int size, unsigned long long id, cmd_opt* opt){
+    char* buffer;
+    int st;
+    off_t sf;
+
+    sf = lseek(*dfw, (size*id), SEEK_SET);
+    buffer = (char*)malloc(size);
+    memset(buffer, 0, size);
+    st = io_all(dfw, buffer, size, 1, opt);
+    log_mesg(1, 0, 0, opt->debug, "write last block%lli, size %i ,status %i\n", id, size, st);
+}
